/*
 * kgraft_patch_bsc966683
 *
 * Fix for CVE-2015-8812, bsc#966683
 *
 *  Upstream commit:
 *  67f1aee6f45059fd6b0f5b0ecb2c97ad0451f6b3
 *
 *  SLE12 commit:
 *  b874ac79cacc7fb95698c2b94ca0ce021446e2bf
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>

#include "local.h"

#define KGR_PATCHED_MODULE	"iw_cxgb3"

static int (*kgr_cxgb3_ofld_send)(struct t3cdev *dev, struct sk_buff *skb);
static int (*kgr_t3_l2t_send_slow)(struct t3cdev *dev, struct sk_buff *skb, struct l2t_entry *e);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "cxgb3_ofld_send", (void *)&kgr_cxgb3_ofld_send },
	{ "t3_l2t_send_slow", (void *)&kgr_t3_l2t_send_slow },
};

static inline int kgr_cxio_fatal_error(struct cxio_rdev *rdev_p)
{
	return rdev_p->flags & CXIO_ERROR_FATAL;
}

static inline int kgr_l2t_send(struct t3cdev *dev, struct sk_buff *skb,
	struct l2t_entry *e)
{
	if (likely(e->state == L2T_STATE_VALID))
		return kgr_cxgb3_ofld_send(dev, skb);
	return kgr_t3_l2t_send_slow(dev, skb, e);
}

/* Patched function */
int kgr_iwch_l2t_send(struct t3cdev *tdev, struct sk_buff *skb,
	struct l2t_entry *l2e)
{
	int	error = 0;
	struct cxio_rdev *rdev;

	rdev = (struct cxio_rdev *)tdev->ulp;
	if (kgr_cxio_fatal_error(rdev)) {
		kfree_skb(skb);
		return -EIO;
	}
	error = kgr_l2t_send(tdev, skb, l2e);
	if (error < 0)
		kfree_skb(skb);
	/*
	 * Fix CVE-2015-8812
	 *  -return error;
	 *  +next row
	 */
	return error < 0 ? error : 0;
}

/* Patched function */
int kgr_iwch_cxgb3_ofld_send(struct t3cdev *tdev, struct sk_buff *skb)
{
	int	error = 0;
	struct cxio_rdev *rdev;

	rdev = (struct cxio_rdev *)tdev->ulp;
	if (kgr_cxio_fatal_error(rdev)) {
		kfree_skb(skb);
		return -EIO;
	}
	error = kgr_cxgb3_ofld_send(tdev, skb);
	if (error < 0)
		kfree_skb(skb);
	/*
	 * Fix CVE-2015-8812
	 *  -return error;
	 *  +next row
	 */
	return error < 0 ? error : 0;
}

static int kgr_patch_bsc966683_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc966683_module_notify(struct notifier_block *nb,
                unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc966683_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc966683_module_nb = {
	.notifier_call = kgr_patch_bsc966683_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc966683_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc966683_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc966683_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc966683_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc966683_module_nb);
}
